
/*
 *  source - $Id: memguard.c,v 1.3 1997/06/10 18:49:27 qian Exp $
 *
 *  a library for memory-based guard tool
 */

/*
#define MG_DEBUG
*/

#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/swap.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/string.h>

#define printf	printk

#include <asm/memguard_machdep.h>
#include <memguard/macros.h>
#include <memguard/types.h>

/*
#include <memguard/main.h>
*/
#include <memguard/memguard.h>
#include "memguard_private.h"

#include <asm/memguard.h>
#include <memguard/memguard_internal.h>
#include <memguard/internal.h>


/*********************************************************************
variables (public)
*********************************************************************/


/*********************************************************************
variables (private)
*********************************************************************/

/* refer to "memguard_private.h" */

static void (*mg_func)(void);
static int mg_errno;


/*********************************************************************
function prototypes (private)
*********************************************************************/

/* refer to "memguard_private.h" */


/*********************************************************************
functions (public)
*********************************************************************/

/*
 * Function   : mgReset
 * Description: reset the tool
 * Parameters : none
 * Returns    : none
 * Plan       : 1. may disable the tool first
 *              2. reset the data structures
 * Note       : the tool must be initialized
 */
void
mgReset()
{
    flags_t flags;

    mg_func = (void (*))mgReset;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. may disable the tool first
     */
    if (mgEnabled) {
	/* enabled, disable the tool */
        mgEnabled = 0;
	mg_disable();
    }

    /*
     * 2. reset the data structures
     */
    mgResetDS();

    __mgSti(flags);
    __mgLeaveCS();
}

/*
 * Function   : mgEnable
 * Description: enable the Tool's memguard page protection fault
 * Parameters : none
 * Returns    : return flag (int)
 *              MG_ENABLE_OK		success
 *              MG_ENABLE_ENABLED	already enabled
 * Plan       : 1. check if tool is already enabled, set it enabled
 *              2. enable the tool
 *              3. return OK
 * Note       : none
 */
int
mgEnable()
{
    flags_t flags;

    mg_func = (void (*))mgEnable;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. check if tool is already enabled, set it enabled
     */
    if (mgEnabled) {
	/* already enabled, return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_ENABLE_ENABLED);
    }
    /* memory guard now disabled, set it enabled */
    mgEnabled = 1;

    /*
     * 2. enable the tool
     */
    mg_enable();

    /*
     * 3. return OK
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_ENABLE_OK);
}

/*
 * Function   : mgDisable
 * Description: disable the Tool's memguard page protection fault
 * Parameters : none
 * Returns    : return flag (int)
 *              MG_DISABLE_OK		success
 *              MG_DISABLE_DISABLED	already disabled
 * Plan       : 1. check if tool is already disabled, set it disabled
 *              2. disable the tool
 *              3. return OK
 * Note       : none
 */
int
mgDisable()
{
    flags_t flags;

    mg_func = (void (*))mgDisable;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. check if tool is already disabled, set it disabled
     */
    if (! mgEnabled) {
	/* already disabled, return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_DISABLE_DISABLED);
    }
    /* memory guard now enabled, set it disabled */
    mgEnabled = 0;

    /*
     * 2. disable the tool
     */
    mg_disable();

    /*
     * 3. return OK
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_DISABLE_OK);
}

/*
 * Function   : mgAdd
 * Description: add an invariant
 * Parameters : name  - invariant name
 *              vaddr - invariant virtual address
 *              size  - invariant size
 * Returns    : return flag (int)
 *              MG_ADD_OK	success
 *              MG_ADD_NAME	invalid name
 *              MG_ADD_VADDR	invalid virtual address
 *              MG_ADD_SIZE	invalid size
 *              MG_ADD_CONFLICT	conflict
 *              MG_ADD_FULL	too many invariants
 * Plan       : 1. check virtual address
 *              2. check invariant size
 *              3. check name length
 *              4. get the corresponding page
 *              5. reserve free invariants
 *              6. check conflict
 *              7. allocate and add it
 *              8. return ok
 */
int
mgAdd(char *name, caddr_t vaddr, u_int size)
{
    register mgpage_t      *pgp;
	/* pointer to an page control block */
    register mginvariant_t *pinvariant;
	/* pointer to an invariant control block */
    register int ix;
    flags_t flags;

    mg_func = (void (*))mgAdd;

    /*
     * 1. check virtual address
     */
    if (vaddr <= (caddr_t)0) {
	/* invalid virtual address, return error */
	return(mg_errno = MG_ADD_VADDR);
    }

    /*
     * 2. check invariant size
     */
    if (size <= 0) {
	/* invalid invariant size, return error */
	return(mg_errno = MG_ADD_SIZE);
    }

    /*
     * 3. check name length
     */
    if (strlen(name) >= MG_SZNAME) {
	/* too long, return error */
	return(mg_errno = MG_ADD_NAME);
    }
    /* valid name */

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 4. get the corresponding page
     */
#ifdef MG_DEBUG
printk("mgAdd: mgGetPage(vaddr=%x, size=%d)\n", (int)vaddr, size);
#endif
    if ((pgp = mgGetPage(vaddr, size)) == NULL) {
	switch (mgErrno) {
	case MG_GETPAGE_VADDR :
	    /* not found, invalid virtual address, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_ADD_VADDR);
	    /* return, no break */
        case MG_GETPAGE_SIZE :
	    /* invalid size, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_ADD_SIZE);
	    /* return, no break */
        }
    }
    /* valid virtual address & size */

    /*
     * 5. reserve free invariants
     */
    for (ix = mgAddrLen2nPage(vaddr, size),
	        pinvariant = mgInvariantsFree;
	    ix > 0 && pinvariant != NULL;
	    ix --, pinvariant = pinvariant->gi_next) {
        ;
    }
    /* check no enough free invariants */
    if (ix > 0) {
        /* no enough free invariants, return error */
        __mgSti(flags);
        __mgLeaveCS();
        return(mg_errno = MG_ADD_FULL);
    }

    /*
     * 6. check conflict
     */
#ifdef MG_DEBUG
printk("mgAdd: mg_checkconflict(pgp=%x, name=%s, vaddr=%x, size=%d\n",
	(int)pgp, name, (int)vaddr, size);
#endif
    if (mg_checkconflict(pgp, name, vaddr, size) != NULL) {
        /* conflict, return error */
        __mgSti(flags);
        __mgLeaveCS();
        return(mg_errno = MG_ADD_CONFLICT);
    }
    /* no conflict */

    /*
     * 7. allocate and add it
     */
#ifdef MG_DEBUG
printk("mg_add(pgp=%x, name=%s, vaddr=%x, size=%d)\n",
	(int)pgp, name, (int)vaddr, size);
#endif
    mg_add(pgp, name, vaddr, size);

    /*
     * 8. return ok
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_ADD_OK);
}

/*
 * Function   : mgDelete
 * Description: delete an invariant
 * Parameters : vaddr - invariant virtual address
 * Returns    : return flag (int)
 *              MG_DELETE_OK	success
 *              MG_DELETE_VADDR	error virtual address: non-exists
 * Plan       : 1. get the corresponding page
 *              2. try to delete the invariant
 *              3. return ok
 */
int
mgDelete(caddr_t vaddr)
{
    flags_t flags;
    register mgpage_t *pgp;

    mg_func = (void (*))mgDelete;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. get the corresponding page
     */
    if ((pgp = mgGetPage(vaddr, 0)) == NULL) {
	/* not found (MG_GETPAGE_SIZE impossible), return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_DELETE_VADDR);
    }

    /*
     * 2. try to delete the invariant
     */
    switch (mg_delete(pgp, vaddr)) {
    case DELETE_SUCCESS :	/* found */
        break;
    case DELETE_VADDR :		/* error virtual address: non-exists */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_DELETE_VADDR);
	/* return, no break */
    }

    /*
     * 3. return ok
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_DELETE_OK);
}

/*
 * Function   : mgSet
 * Description: set an invariant
 * Parameters : vaddr   - invariant virtual address
 *              size    - invariant size
 *              content - invariant content
 * Returns    : return flag (int)
 *              MG_SET_OK	success
 *              MG_SET_VADDR	error virtual address: non-exists
 *              MG_SET_SIZE	error size
 *              MG_SET_CONTENT	error content: copy
 * Plan       : 1. get the corresponding page
 *              2. search the invariant
 *              3. set the invariant
 *              4. return ok
 */
int
mgSet(caddr_t vaddr, u_int size, char *content)
{
    flags_t flags;
    register mgpage_t *pgp;

    mg_func = (void (*))mgSet;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. get the corresponding page
     */
    if ((pgp = mgGetPage(vaddr, size)) == NULL) {
	switch (mgErrno) {
	case MG_GETPAGE_VADDR :
	    /* not found, invalid virtual address, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_SET_VADDR);
	    /* return, no break */
        case MG_GETPAGE_SIZE :
	    /* invalid size, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_SET_SIZE);
	    /* return, no break */
        }
    }
    /* valid virtual address & size */

    /*
     * 2. search the invariant
     */
    switch (mg_search(pgp, vaddr, size)) {
    case SEARCH_SUCCESS :	/* found */
	break;
    case SEARCH_VADDR :		/* error virtual address: non-exists */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_SET_VADDR);
	/* return, no break */
    case SEARCH_SIZE :		/* error size */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_SET_SIZE);
	/* return, no break */
    }

    /*
     * 3. set the invariant
     */
    if (mg_sga(SGA_SET, pgp, vaddr, size, content) != SGA_OK) {
	/* set content error, return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_SET_CONTENT);
    }

    /*
     * 4. return ok
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_SET_OK);
}

/*
 * Function   : mgGet
 * Description: get an invariant
 * Parameters : vaddr   - invariant virtual address
 *              size    - invariant size
 *              content - invariant content
 * Returns    : return flag (int)
 *              MG_GET_OK	success
 *              MG_GET_VADDR	error virtual address: non-exists
 *              MG_GET_SIZE	error size
 *              MG_GET_CONTENT	error content: copy
 * Plan       : 1. get the corresponding page
 *              2. search the invariant
 *              3. set the invariant
 *              4. return ok
 */
int
mgGet(caddr_t vaddr, u_int size, char *content)
{
    flags_t flags;
    register mgpage_t *pgp;

    mg_func = (void (*))mgGet;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. get the corresponding page
     */
    if ((pgp = mgGetPage(vaddr, size)) == NULL) {
	switch (mgErrno) {
	case MG_GETPAGE_VADDR :
	    /* not found, invalid virtual address, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_GET_VADDR);
	    /* return, no break */
        case MG_GETPAGE_SIZE :
	    /* invalid size, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_GET_SIZE);
	    /* return, no break */
        }
    }
    /* valid virtual address & size */

    /*
     * 2. search the invariant
     */
    switch (mg_search(pgp, vaddr, size)) {
    case SEARCH_SUCCESS :	/* found */
	break;
    case SEARCH_VADDR :		/* error virtual address: non-exists */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_GET_VADDR);
	/* return, no break */
    case SEARCH_SIZE :		/* error size */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_GET_SIZE);
	/* return, no break */
    }

    /*
     * 3. get the invariant
     */
    if (mg_sga(SGA_GET, pgp, vaddr, size, content)
	    != SGA_OK) {
	/* get content error, return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_GET_CONTENT);
    }

    /*
     * 4. return ok
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_GET_OK);
}

/*
 * Function   : mgAssert
 * Description: assert an invariant
 * Parameters : vaddr   - invariant virtual address
 *              size    - invariant size
 *              content - invariant content
 * Returns    : return flag (int)
 *              MG_ASSERT_OK	success
 *              MG_ASSERT_VADDR	error virtual address: non-exists
 *              MG_ASSERT_SIZE	error size
 *              MG_ASSERT_CONTENT	error content: copy
 * Plan       : 1. get the corresponding page
 *              2. search the invariant
 *              3. set the invariant
 *              4. return ok
 */
int
mgAssert(caddr_t vaddr, u_int size, char *content)
{
    flags_t flags;
    register mgpage_t *pgp;

    mg_func = (void (*))mgAssert;

    __mgEnterCS();
    __mgCli(flags);

    /*
     * 1. get the corresponding page
     */
    if ((pgp = mgGetPage(vaddr, size)) == NULL) {
	switch (mgErrno) {
	case MG_GETPAGE_VADDR :
	    /* not found, invalid virtual address, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_ASSERT_VADDR);
	    /* return, no break */
        case MG_GETPAGE_SIZE :
	    /* invalid size, return error */
            __mgSti(flags);
            __mgLeaveCS();
	    return(mg_errno = MG_ASSERT_SIZE);
	    /* return, no break */
        }
    }
    /* valid virtual address & size */

    /*
     * 2. search the invariant
     */
    switch (mg_search(pgp, vaddr, size)) {
    case SEARCH_SUCCESS :	/* found */
	break;
    case SEARCH_VADDR :		/* error virtual address: non-exists */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_ASSERT_VADDR);
	/* return, no break */
    case SEARCH_SIZE :		/* error size */
	/* return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_ASSERT_SIZE);
	/* return, no break */
    }

    /*
     * 3. assert the invariant
     */
    if (mg_sga(SGA_ASSERT, pgp, vaddr, size, content) != SGA_OK) {
	/* assert content error, return error */
        __mgSti(flags);
        __mgLeaveCS();
	return(mg_errno = MG_ASSERT_CONTENT);
    }

    /*
     * 4. return ok
     */
    __mgSti(flags);
    __mgLeaveCS();
    return(mg_errno = MG_ASSERT_OK);
}

/*
 * Function   : mgPrintError
 * Description: print an error message
 * Parameters : none
 * Returns    : return flag (int)
 *              MG_PERROR_OK	success
 *              MG_PERROR_ERROR	error
 * Plan       : 1. check the function pointer, print an error message
 *              2. return
 */
int
mgPrintError()
{
    int errflag = 0;

    /*
     * 1. check the function pointer, print an error message
     */
    if (mg_func == (void (*))mgReset) {
	    printf("mgReset(): Successfully\n");
    } else if (mg_func == (void (*))mgEnable) {
        switch (mg_errno) {
        case MG_ENABLE_OK :
	    printf("mgEnable(): Successfully\n");
            break;
        case MG_ENABLE_ENABLED :
	    printf("mgEnable(): Already enabled\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgDisable) {
        switch (mg_errno) {
        case MG_DISABLE_OK :
	    printf("mgDisable(): Successfully\n");
            break;
        case MG_DISABLE_DISABLED :
	    printf("mgDisable(): Already disabled\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgAdd) {
        switch (mg_errno) {
        case MG_ADD_OK :
	    printf("mgAdd(): Successfully\n");
            break;
        case MG_ADD_VADDR :
	    printf("mgAdd(): Invaid virtual address\n");
            break;
        case MG_ADD_SIZE :
	    printf("mgAdd(): Invaid size\n");
            break;
        case MG_ADD_NAME :
	    printf("mgAdd(): Invaid name\n");
            break;
        case MG_ADD_CONFLICT :
	    printf("mgAdd(): Conflict\n");
            break;
        case MG_ADD_FULL :
	    printf("mgAdd(): Too many invariants\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgDelete) {
        switch (mg_errno) {
        case MG_DELETE_OK :
	    printf("mgDelete(): Successfully\n");
            break;
        case MG_DELETE_VADDR :
	    printf("mgDelete(): Doesn't exist\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgSet) {
        switch (mg_errno) {
        case MG_SET_OK :
	    printf("mgSet(): Successfully\n");
            break;
        case MG_SET_VADDR :
	    printf("mgSet(): Doesn't exist\n");
            break;
        case MG_SET_SIZE :
	    printf("mgSet(): Invalid size\n");
            break;
        case MG_SET_CONTENT :
	    printf("mgSet(): Copy failed\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgGet) {
        switch (mg_errno) {
        case MG_GET_OK :
	    printf("mgGet(): Successfully\n");
            break;
        case MG_GET_VADDR :
	    printf("mgGet(): Doesn't exist\n");
            break;
        case MG_GET_SIZE :
	    printf("mgGet(): Invalid size\n");
            break;
        case MG_GET_CONTENT :
	    printf("mgGet(): Copy failed\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgAssert) {
        switch (mg_errno) {
        case MG_ASSERT_OK :
	    printf("mgAssert(): Successfully\n");
            break;
        case MG_ASSERT_VADDR :
	    printf("mgAssert(): Doesn't exist\n");
            break;
        case MG_ASSERT_SIZE :
	    printf("mgAssert(): Invalid size\n");
            break;
        case MG_ASSERT_CONTENT :
	    printf("mgAssert(): Unmatched contents\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else if (mg_func == (void (*))mgPrintError) {
        switch (mg_errno) {
        case MG_PERROR_OK :
	    printf("mgPrintError(): Successfully\n");
            break;
        case MG_PERROR_ERROR :
	    printf("mgPrintError(): Error\n");
            break;
        default :
            errflag = 1;
            break;
        }
    } else {
        errflag = 1;
    }

    /*
     * 2. return
     */
    /* mg_func cannot be assigned before here */
    mg_func = (void (*))mgPrintError;
    return(mg_errno = (errflag == 0 ? MG_PERROR_OK : MG_PERROR_ERROR));
}


/*********************************************************************
functions (private)
*********************************************************************/

/* refer to "memguard_private.c" */

